/*
 *rmpInverse.c
 *
 * Re-Created on: 14 Dec 2008
 *    Author: rchuso ( see mjollnir.com )
 */

#include "maths/rmpInverse.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>

static const float ROUNDOFF_ERROR = (float) 1.0E-6;

static inline float absoluteValue( const float a ) {
    return (a < 0.0) ? -a : a;
}

/**
 * initI
 * Initialize this matrix to the Identity matrix
 */
static void RMP_initI( float* const * const M, // matrix (square)
                   const int rows,
                   const int cols, // matrix size (rows == cols)
                   int threaded ) { // attempt to use the omp threading
    int row;
    for( row = 0; row < rows; ++row ) {
        int col;
        for( col = 0; col < cols; ++col ) {
            M[row][col] = (float) 0.0;
        }
        M[row][row] = (float) 1.0;
    }
}
/**
 * innerMultiply
 * Multiply the first matrix times the second, but only as far as the innerP
 */
static void RMP_innerMultiply( float* const * const G, // left matrix
                           float* const * const H, // right matrix
                           const int rows,
                           const int innerP,
                           const int cols, // ( rows, innerP ) * ( innerP, cols )
                           float* const * const Result, // result of calculation
                           int threaded ) { // use omp threading
    int row;
    for( row = 0; row < rows; ++row ) {
        int col;
        for( col = 0; col < cols; ++col ) {
            float temp = G[row][0] * H[0][col];
            int inner;
            for( inner = 1; inner < innerP; ++inner ) {
                temp += G[row][inner] * H[inner][col];
            }
            Result[row][col] = temp;
        }
    }
}
/**
 * multiplyColAddToCol
 * Multiply the prod times the colB and add this to the element in the same row of colA
 */
static inline void RMP_multiplyColAddToCol( float* const * const M, // matrix
                                        const int rows, // matrix rows (cols isn't needed)
                                        const int pivot, // beginning diagonal position
                                        const int colA, // destination col
                                        const float prod, // amount to adjust
                                        const int colB ) { // source col
    int row;
    for( row = pivot; row < rows; ++row ) {
        M[row][colA] += (prod * M[row][colB]);
    }
}
/**
 * multiplyRowAddToColumn
 * Multiply the prod times the rowB and add this to the element in the same column of rowA
 */
static inline void RMP_multiplyRowAddToRow( float* const * const M, // matrix
                                        const int cols, // matrix cols (rows isn't needed)
                                        const int pivot, // beginning diagonal position
                                        const int rowA, // destination row
                                        const float prod, // amount of rowB to add
                                        int rowB ) { // source row
    int col;
    for( col = pivot; col < cols; ++col ) {
        M[rowA][col] += (prod * M[rowB][col]);
    }
}
/**
 * multiplyRow
 * Multiply this value times every entry in the row
 */
static inline void RMP_multiplyRow( const int pivot, // starting diagonal position
                                const int row, // row to scale
                                float* const * const M, // matrix
                                const int cols, // cols in matrix
                                const float prod ) { // amount to scale
    int col;
    for( col = pivot; col < cols; ++col ) {
        M[row][col] *= prod;
    }
}
/**
 * reduceUsingPivot
 * Use Rand's method of reducing the matrix.
 */
static void RMP_reduceUsingPivot( const int pivot, // current diagonal number
                              float* const * const A, // source matrix
                              float* const * const S, // column operations matrix
                              float* const * const T, // row operations matrix
                              const int rows,
                              const int cols, // size of the A matrix
                              int threaded ) { // flag to use omp threading
    int row, col;
    float prod = 1.0 / A[pivot][pivot]; // already checked that it's not zero
    RMP_multiplyRow( pivot, pivot, A, cols, prod );
    A[pivot][pivot] = 1.0; // remove any roundoff errors
    RMP_multiplyRow( 0, pivot, T, rows, prod );

    for( row = 1 + pivot; row < rows; ++row ) {
        prod = -A[row][pivot];
        RMP_multiplyRowAddToRow( A, cols, pivot, row, prod, pivot );
        RMP_multiplyRowAddToRow( T, rows, 0, row, prod, pivot );
    }

    for( col = 1 + pivot; col < cols; ++col ) {
        prod = -A[pivot][col];
        RMP_multiplyColAddToCol( A, rows, pivot, col, prod, pivot );
        RMP_multiplyColAddToCol( S, cols, 0, col, prod, pivot );
    }
}
/**
 * swapCols
 * Swap the columns in the supplied matrix from the pivot to the end of the row
 */
static inline void RMP_swapCols( float* const * const M, const int rows, const int cols, const int pivot, const int colA, const int colB ) {
    float temp;
    int row;
    for( row = pivot; row < rows; ++row ) {
        temp = M[row][colA];
        M[row][colA] = M[row][colB];
        M[row][colB] = temp;
    }

}
/**
 * swapRows
 * Swap the rows in the supplied matrix from the pivot to the end of the cols
 */
static inline void RMP_swapRows( float* const * const M, const int cols, const int pivot, const int rowA, const int rowB ) {
    float temp;
    int col;
    for( col = pivot; col < cols; ++col ) {
        temp = M[rowA][col];
        M[rowA][col] = M[rowB][col];
        M[rowB][col] = temp;
    }
}
/**
 * swapInNextPivot
 * Searches through the remaining matrix for the absolute largest value to use as the next pivot and swaps it into position
 */
static int RMP_swapInNextPivot( const int rank, //
                            float* const * const A, // matrix being reduced
                            float* const * const S, // column operations workspace
                            float* const * const T, // row operations workspace
                            const int rows,
                            const int cols, // size of the A matrix
                            int threaded ) { // attempt to use omp threading
    int row, col;
    int maxRow = rank;
    int maxCol = rank;
    register float temp;
    float maxValue = absoluteValue( A[rank][rank] );
    for( row = rank; row < rows; ++row ) {
        for( col = rank; col < cols; ++col ) {
            temp = absoluteValue( A[row][col] );
            if( maxValue < temp ) {
                maxValue = temp;
                maxRow = row;
                maxCol = col;
            }
        }
    }
    if( rank != maxRow ) {
        {
            RMP_swapRows( A, cols, rank, rank, maxRow );
            RMP_swapRows( T, rows, 0, rank, maxRow );
        }
    }
    if( rank != maxCol ) {
        {
            RMP_swapCols( A, rows, cols, rank, rank, maxCol );
            RMP_swapCols( S, cols, cols, 0, rank, maxCol );
        }
    }
    return (ROUNDOFF_ERROR < fabs( A[maxRow][maxCol] )); // wish I didn't have to do this
}
/**
 * rankReduce
 * Perform row and column operations on the A, S, and T, this is destructive to A.
 */
static int RMP_rankReduce( float* const * const A, float* const * const S, float* const * const T, const int rows, const int cols, int threaded ) {
    int minRC = (rows < cols ? rows : cols);
    int rank = 0;
    int diagonal;
    for( diagonal = 0; diagonal < minRC; ++diagonal ) {
        if( !RMP_swapInNextPivot( rank, A, S, T, rows, cols, threaded ) ) return rank;
        RMP_reduceUsingPivot( rank, A, S, T, rows, cols, threaded );
        ++rank;
    }
    return rank;
}
/**
 * getMatrixDuplicate
 * Create a duplicate matrix - does not allocate memory
 */
static void RMP_getMatrixDuplicate( float* const * const Copy, float* const * const M, const int rows, const int cols ) {
    int row, col;
    for( row = 0; row < rows; ++row ) {
        for( col = 0; col < cols; ++col ) {
            Copy[row][col] = M[row][col];
        }
    }
}
/**
 * getATransposeA
 * Calculate the A<sup>T</sup> * A
 * This is a symmetric product.
 */
static void RMP_getATransposeA( float* const * const A, // matrix - not transposed
                            const int rows,
                            const int cols, // matrix size
                            float* const * const Result, // result matrix (cols,cols)
                            int threaded ) { // attempt to use omp
    int col;
    // this can't be included because we call this from Fortran, our f**king application forces
    // everything to be on the heap... but openmp requires it be on the stack.
    //#pragma omp parallel for private( col ) if( threaded )
    for( col = 0; col < cols; ++col ) {
        int row;
        for( row = col; row < cols; ++row ) {
            float temp;
            temp = A[0][row] * A[0][col];
            int inner;
            for( inner = 1; inner < rows; ++inner ) {
                temp += A[inner][row] * A[inner][col];
            }
            Result[row][col] = Result[col][row] = temp;
        }
    }
}
/**
 * getAATranspose
 * Calculate the A * A<sup>T</sup>
 * This is a symmetric product.
 */
static void RMP_getAATranspose( float* const * const A, // original matrix
                            const int rows,
                            const int cols, // size of matrix
                            float* const * const Result, // result matrix
                            int threaded ) { // attempt to use omp
    int row;
    // this can't be included because we call this from Fortran, our f**king application forces
    // everything to be on the heap... but openmp requires it be on the stack.
    //#pragma omp parallel for private( row ) if( threaded )
    for( row = 0; row < rows; ++row ) {
        int col;
        for( col = row; col < rows; ++col ) {
            float temp;
            temp = A[row][0] * A[col][0];
            int inner;
            for( inner = 1; inner < cols; ++inner ) {
                temp += A[row][inner] * A[col][inner];
            }
            Result[row][col] = Result[col][row] = temp;
        }
    }
}
/**
 * multiplyInverseATranspose
 * Multiply the B matrix times the A-Transpose. The A is supplied instead of the A-Transpose.
 */
static void RMP_multiplyInverseATranspose( float* const * const B, // B matrix (cols,cols)
                                       float* const * const A, // original A matrix (rows,cols)
                                       const int rows,
                                       const int cols, // size of original A matrix
                                       float* const * const Result, // result matrix
                                       int threaded ) {
    int row;
    for( row = 0; row < rows; ++row ) {
        int col;
        for( col = 0; col < cols; ++col ) {
            int inner;
            float temp = B[col][0] * A[row][0];
            for( inner = 1; inner < cols; ++inner ) {
                temp += B[col][inner] * A[row][inner];
            }
            Result[col][row] = temp;
        }
    }
}
/**
 * multiplyATransposeInverse
 * Multiply the A-Transpose matrix times the B matrix. The A is supplied instead of the A-Transpose.
 */
static void RMP_multiplyATransposeInverse( float* const * const A, // original A matrix (rows,cols)
                                       const int rows,
                                       const int cols, // size of A matrix
                                       float* const * const B, // B matrix (rows,rows)
                                       float* const * const Result, // result matrix
                                       int threaded ) { // attempt to use the omp for threading
    int col;
    for( col = 0; col < cols; ++col ) {
        int row;
        for( row = 0; row < rows; ++row ) {
            int inner;
            float temp = A[0][col] * B[0][row];
            for( inner = 1; inner < rows; ++inner ) {
                temp += A[inner][col] * B[inner][row];
            }
            Result[col][row] = temp;
        }
    }
}

typedef enum {
    OVER_SPECIFIED,
    SQUARE_SPECIFIED,
    UNDER_SPECIFIED
} Specified;

/**
 * allocateMatrixMemory
 * Allocate memory for the workspace
 */
int RMP_allocateMatrixMemory( float*** M, const int rows, const int cols ) {
    int row;
    *M = malloc( rows * sizeof(float*) );
    if( NULL == *M ) return 0;
    for( row = 0; row < rows; ++row ) {
        ( *M)[row] = malloc( cols * sizeof(float*) );
        if( NULL == ( *M)[row] ) return 0;
    }
    return 1;
}
/**
 * freeMatrixMemory
 * Free the allocated memory
 */
int RMP_freeMatrixMemory( float*** M, const int rows ) {
    int row;
    if( NULL != ( *M) ) {
        for( row = 0; row < rows; ++row ) {
            if( NULL != ( *M)[row] ) free( ( *M)[row] );
            else return 0;
        }
        free( *M );
    }
    ( *M) = NULL;
    return 1;
}

/**
 * rmpInverseW
 * Invert the A matrix using the supplied workspaces for S and T.
 * Call this one if you intend to manage the workspaces.
 */
int RMP_inverseW( float* const * const A, //  original
                 float* const * const S,
                 float* const * const T, // workspaces
                 const int rows,
                 const int cols, // size of A
                 float* const * const Result, // result of the inversion
                 int threaded ) { // use omp threading
    int innerP;
    RMP_initI( S, cols, cols, threaded );
    RMP_initI( T, rows, rows, threaded );
    innerP = RMP_rankReduce( A, S, T, rows, cols, threaded );
    RMP_innerMultiply( S, T, cols, innerP, rows, Result, threaded );
    return 1;
}

/**
 * rmpInverseNd
 * Non-destructive version of rmpInverse.
 */
int RMP_inverseNd( float* const * const A, // original matrix
                  const int rows,
                  const int cols, // size of original
                  float* const * const Result, // result of inversion
                  int threaded ) { // use omp threading
    int response = 0;
    float**S;
    float**T;
    float**Copy;
    if( RMP_allocateMatrixMemory( &S, cols, cols ) ) {
        if( RMP_allocateMatrixMemory( &T, rows, rows ) ) {
            if( RMP_allocateMatrixMemory( &Copy, rows, cols ) ) {
                RMP_getMatrixDuplicate( Copy, A, rows, cols );
                response = RMP_inverseW( Copy, S, T, rows, cols, Result, threaded );
                RMP_freeMatrixMemory( &Copy, rows );
            }
            RMP_freeMatrixMemory( &T, rows );
        }
        RMP_freeMatrixMemory( &S, cols );
    }
    return response;
}
/**
 * rmpInverse
 * Invert the supplied matrix A by allocating workspaces and calling the above rmpInverseW routine
 */
int RMP_inverse( float* const * const A, int rows, int cols, float* const * const Result, int threaded ) {
    int response;
    float**S;
    float**T;
    if( RMP_allocateMatrixMemory( &S, cols, cols ) ) {
        if( RMP_allocateMatrixMemory( &T, rows, rows ) ) {
            response = RMP_inverseW( A, S, T, rows, cols, Result, threaded );
            RMP_freeMatrixMemory( &T, rows );
        }
        RMP_freeMatrixMemory( &S, cols );
    }
    return response;
}

/**
 * rmpInverseMPW
 * Invert the matrix using Moore-Penrose as the base, but performing the inverse itself with the above inversion
 * routine which doesn't suffer from traditional inversion methods.
 * This routine requires all the temporary workspaces - call this one if you intend to manage them.
 */
static int RMP_inverseMPW( float* const * const A, // original - is destroyed
                          const int rows,
                          const int cols, // size of original
                          float* const * const Result, // the inverse
                          const int minRowCol, // sizes for S and T
                          float* const * const S, // S workspace ( minRowCol, minRowCol )
                          float* const * const T, // T workspace ( minRowCol, minRowCol )
                          float* const * const WorkSquareSmall, // WorkSquareSmall A<sup>T</sup>A or AA<sup>T</sup> ( minRowCol, minRowCol )
                          float* const * const WorkSquareSmallInverted, // Inverted matrix workspace before final product with A<sup>T</sup> ( right or left )
                          int threaded ) { // attempt to use the omp library for threading
    //
    int response;
    Specified specified = (cols > rows ? UNDER_SPECIFIED : (cols < rows ? OVER_SPECIFIED : SQUARE_SPECIFIED));
    switch( specified ) {
        case OVER_SPECIFIED:
            RMP_getATransposeA( A, rows, cols, WorkSquareSmall, threaded );
            response = RMP_inverseW( WorkSquareSmall, S, T, cols, cols, WorkSquareSmallInverted, threaded );
            RMP_multiplyInverseATranspose( WorkSquareSmallInverted, A, rows, cols, Result, threaded );
            break;
        case SQUARE_SPECIFIED:
        case UNDER_SPECIFIED:
            RMP_getAATranspose( A, rows, cols, WorkSquareSmall, threaded );
            response = RMP_inverseW( WorkSquareSmall, S, T, rows, rows, WorkSquareSmallInverted, threaded );
            RMP_multiplyATransposeInverse( A, rows, cols, WorkSquareSmallInverted, Result, threaded );
            break;
    }
    return response;
}
/**
 * rmpInverseMP
 * Allocates workspaces and calls the rmpInverseMPW to perform the inverse.
 */
int RMP_inverseMP( float* const * const A, // Matrix to invert ( rows, cols )
                  const int rows,
                  const int cols, // Input A Matrix size
                  float* const * const Result, // Result of the inverse operation ( cols, rows )
                  int threaded ) { // attempt to use omp library for threading
    int minRowCol = (rows < cols ? rows : cols);
    int response = 0;
    float**S;
    float**T;
    float**WorkSquareSmall;
    float**WorkSquareSmallInverted;
    if( RMP_allocateMatrixMemory( &S, minRowCol, minRowCol ) ) {
        if( RMP_allocateMatrixMemory( &T, minRowCol, minRowCol ) ) {
            if( RMP_allocateMatrixMemory( &WorkSquareSmall, minRowCol, minRowCol ) ) {
                if( RMP_allocateMatrixMemory( &WorkSquareSmallInverted, minRowCol, minRowCol ) ) {
                    response = RMP_inverseMPW( A, rows, cols, Result, minRowCol, S, T, WorkSquareSmall, WorkSquareSmallInverted, threaded );
                    RMP_freeMatrixMemory( &WorkSquareSmallInverted, minRowCol );
                }
                RMP_freeMatrixMemory( &WorkSquareSmall, minRowCol );
            }
            RMP_freeMatrixMemory( &T, minRowCol );
        }
        RMP_freeMatrixMemory( &S, minRowCol );
    }
    return response;
}
/**
 * rmpInverseNdMP
 * Non-Destructive version of the rmpInverseMP routine - allocates memory for a
 * copy of the supplied matrix and performs the inverse.
 */
int RMP_inverseNdMP( float* const * const A, // Matrix to invert ( rows, cols )
                    const int rows,
                    const int cols, // Input A Matrix size
                    float* const * const Result, // Result of the inverse operation ( cols, rows )
                    int threaded ) { // attempt to use omp library for threading
    int minRowCol = (rows < cols ? rows : cols);
    int response;
    float**S;
    float**T;
    float**WorkSquareSmall;
    float**WorkSquareSmallInverted;
    float**Copy;
    if( RMP_allocateMatrixMemory( &S, minRowCol, minRowCol ) ) {
        if( RMP_allocateMatrixMemory( &T, minRowCol, minRowCol ) ) {
            if( RMP_allocateMatrixMemory( &WorkSquareSmall, minRowCol, minRowCol ) ) {
                if( RMP_allocateMatrixMemory( &WorkSquareSmallInverted, minRowCol, minRowCol ) ) {
                    if( RMP_allocateMatrixMemory( &Copy, rows, cols ) ) {
                        RMP_getMatrixDuplicate( Copy, A, rows, cols );
                        response = RMP_inverseMPW( Copy, rows, cols, Result, minRowCol, S, T, WorkSquareSmall, WorkSquareSmallInverted, threaded );
                        RMP_freeMatrixMemory( &Copy, rows );
                    }
                    RMP_freeMatrixMemory( &WorkSquareSmallInverted, minRowCol );
                }
                RMP_freeMatrixMemory( &WorkSquareSmall, minRowCol );
            }
            RMP_freeMatrixMemory( &T, minRowCol );
        }
        RMP_freeMatrixMemory( &S, minRowCol );
    }
    return response;
}
