/*
 *rmpInverse.h
 *
 * Re-Created on: 14 Dec 2008
 *    Author: rchuso (see mjollnir.com)
 *
 */

#ifndef rmpInverse_H_
#define rmpInverse_H_

/**
 * Allocate and free memory.
 * Memory in the array is accessed directly:
 *     M[2][3] = 14.5;
 * and these routines allocate and free the memory.
 * typical use:
 *     float ** A;
 *     RMP_allocateMatrixMemory( &A, rows, cols );
 *     ...
 *     RMP_freeMatrixMemory( &A, rows );
 */
int RMP_allocateMatrixMemory( // allocate matrix memory
                          float*** M, // Matrix - pointer to two-dimensional array
                          const int rows,
                          const int cols ); // Matrix size
int RMP_freeMatrixMemory( // free matrix memory allocated by the above routine
                      float*** M, // Matrix
                      const int rows ); // Matrix size

/**
 * rmpInverseW
 * Main inverse method - called by all the other matrix inversion routines.
 * Destroys the supplied A matrix.
 * This routine performs row and column operations to calculate the inverse,
 * and uses the supplied S and T workspaces.
 * Result contains the inverted routine with size Result(cols,rows)
 * For the operation of this function consult http://mjollnir.com
 */
int RMP_inverseW( float* const * const A, // Original matrix
                 float* const * const S, // S workspace
                 float* const * const T, // T workspace
                 const int rows,
                 const int cols, // Original matrix size
                 float* const * const Result, // Inverted matrix
                 int threaded ); // use omp threading (see note at bottom)
/**
 * rmpInverseNd
 * Non-Destructive matrix inversion - allocates memory for a copy of the A
 * matrix, inverts the copy using the rmpInverseW routine, destroys the copy.
 */
int RMP_inverseNd( float* const * const A, // Original matrix
                  const int rows,
                  const int cols, // Original matrix size
                  float* const * const Result, // Inverted matrix
                  int threaded ); // use omp threading (see note at bottom)
/**
 * rmpInverse
 * Destructive matrix inversion - allocates memory for the workspaces and calls
 * rmpInverseW to do the actual inverse, then performs cleanup on the allocated
 * memory.
 */
int RMP_inverse( float* const * const A, // Original matrix
                int rows,
                int cols, // Original matrix size
                float* const * const Result, // Inverted matrix
                int threaded ); // use omp threading (see note at bottom)
/**
 * rmpInverseMP
 * Destructive matrix inversion using a combination of the Moore-Penrose method
 * and Rand's matrix inversion routine.
 *   A<sup>-1</sup> = (A<sup>T</sup> * A)<sup>-1</sup> * A<sup>T</sup>
 *   or
 *   A<sup>-1</sup> = A<sup>T</sup>(A * A<sup>T</sup>)<sup>-1</sup>
 * This routine allocates workspaces and calls rmpInverseMPW to do the actual
 * matrix inverse.
 */
int RMP_inverseMP( float* const * const A, // Matrix to invert (rows,cols)
                  const int rows,
                  const int cols, // Original matrix sizes
                  float* const * const Result, // Inverted matrix
                  int threaded ); // attempt to use omp (see note at bottom)
/**
 * rmpInverseNdMP
 * Non-Destructive version of rmpInverseMP.
 * Allocates memory for a copy of the A matrix, performs the inverse, removes
 * allocated workspaces, and returns the Result.
 */
int RMP_inverseNdMP( float* const * const A, // Matrix to invert (rows,cols)
                    const int rows,
                    const int cols, // Original matrix sizes
                    float* const * const Result, // Inverted matrix
                    int threaded ); // attempt to use omp (see note at bottom)

// NOTE: the "threaded" variable is used to allow the routine to use OpenMP library calls.
// Unfortunately, this is linked from applications written in Fortran, and our compilation
// puts data on the heap instead of the stack. This is in violation of what OpenMP needs.
// Therefore, this variable is currently not being used.

#endif /* rmpInverse_H_ */
